/*
 * shohaku
 * Copyright (C) 2005  tomoya nagatani
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package shohaku.core.util.cel;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import shohaku.core.collections.Parameters;
import shohaku.core.util.DateUtils;

/**
 * 日付を生成する式を提供します。
 */
public class DateTimeExpression implements Expression {

    /* 有効な日付フォーマットの配列。 */
    private static final String[] dateFormats;
    static {
        List list = new ArrayList();
        list.add("yyyy");
        list.add("yyyy-MM");
        list.add("yyyy-MM-dd");
        list.add("yyyy-MM-dd HH");
        list.add("yyyy-MM-dd HH:mm");
        list.add("yyyy-MM-dd HH:mm:ss");
        list.add("yyyy-MM-dd HH:mm:ss.S");
        list.add("yyyy-MM-dd HH:mm:ss.SS");
        list.add("yyyy-MM-dd HH:mm:ss.SSS");
        list.add("HH:mm");
        list.add("HH:mm:ss");
        list.add("HH:mm:ss.S");
        list.add("HH:mm:ss.SS");
        list.add("HH:mm:ss.SSS");
        dateFormats = (String[]) list.toArray(new String[0]);
    }

    /* この式のリテラル文字を定義します。 */
    private static final Literal[] literals = new Literal[] { new Literal('|', '|') };

    /**
     * リテラル文字 '|', '|' を返します。
     * 
     * @return リテラル文字
     */
    public Literal[] getLiteral() {
        return literals;
    }

    /**
     * 式を実行して値を返します。
     * 
     * @param binder
     *            処理基のバインダー
     * @param literal
     *            リテラル文字
     * @param expression
     *            式
     * @param begin
     *            開始インデックス
     * @param end
     *            終了インデックス
     * @param values
     *            引数値の一覧
     * @return 式の結果
     */
    public Object execute(CELBinder binder, Literal literal, String expression,
            int begin, int end, Parameters values) {
        Date d = DateUtils.parse(expression.substring(begin, end), dateFormats);
        if (d == null) {
            throw new IllegalArgumentException();
        }
        return d;
    }

    /**
     * 式の終端のインデックスを検索して返します。 発見出来ない場合は -1 を返します。
     * 
     * @param binder
     *            処理基のバインダー
     * @param expression
     *            式
     * @param literal
     *            リテラル文字
     * @param begin
     *            開始インデックス
     * @return 式の終端のインデックス
     */
    public int findNextIndex(CELBinder binder, Literal literal, String expression, int begin) {
        int end = (begin + 1);//'|'.length() == 1
        while (end < expression.length()) {
            end = expression.indexOf('|', end);
            if (-1 == end) {
                return -1;
            }
            if (0 == end || (0 < end && '\\' != expression.charAt(end - 1))) {
                break;
            }
            end++;
        }
        return (end + 1);//end to next
    }
}