/*
 * shohaku
 * Copyright (C) 2005  tomoya nagatani
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */
package shohaku.core.util.cel;

import java.util.ArrayList;

import shohaku.core.beans.BeanUtilities;
import shohaku.core.beans.InvocationBeansException;
import shohaku.core.collections.Parameters;
import shohaku.core.lang.NoSuchResourceException;
import shohaku.core.lang.ResourceLoader;
import shohaku.core.lang.Seek;

/**
 * コンストラクタまたはスタティックメソッドからオブジェクトインスタンスを生成する式を提供します。
 */
public class ObjectCreateExpression implements Expression {

    /* この式のリテラル文字を定義します。 */
    private static final Literal[] literals = new Literal[] { new Literal('@', '@') };

    /**
     * リテラル文字 '@', '@' を返します。
     * 
     * @return リテラル文字
     */
    public Literal[] getLiteral() {
        return literals;
    }

    /**
     * 式を実行して値を返します。
     * 
     * @param binder
     *            処理基のバインダー
     * @param literal
     *            リテラル文字
     * @param expression
     *            式
     * @param begin
     *            開始インデックス
     * @param end
     *            終了インデックス
     * @param values
     *            引数値の一覧
     * @return 式の結果
     */
    public Object execute(CELBinder binder, Literal literal, String expression, int begin, int end, Parameters values) {

        String exp = expression.substring(begin, end).trim();

        int off = 0;
        Class c = null;
        String className = null;
        String methodName = null;
        int methodLiteral = exp.indexOf('#');
        if (-1 < methodLiteral) {
            className = exp.substring(0, methodLiteral);
            int argStart = exp.indexOf('(', methodLiteral);
            if (-1 < argStart) {
                methodName = exp.substring((methodLiteral + 1), argStart);
                off = argStart + 1;
            } else {
                //err
            }
        } else {
            int argStart = exp.indexOf('(');
            if (-1 < argStart) {
                className = exp.substring(0, argStart);
                off = argStart + 1;
            } else {
                //err
            }
        }
        try {
            c = ResourceLoader.getClass(className);
        } catch (NoSuchResourceException e) {
            throw new CELFormatException("bean invocation err :" + exp, e);
        }

        Object[] args = null;
        if (')' != exp.charAt(off)) {
            ArrayList list = new ArrayList();
            while (off < exp.length()) {
                int inx = binder.findNextIndex(exp, off);
                if (-1 < inx) {
                    list.add(binder.getValue(exp, off, inx, values));
                    off = Seek.skipTo(exp, ' ', inx);
                    if (')' == exp.charAt(off)) {
                        break;
                    }
                    if (',' != exp.charAt(off)) {
                        throw new IllegalArgumentException();
                    }
                    off = Seek.skipTo(exp, ' ', off + 1);
                } else {
                    break;
                }
            }
            if(list.isEmpty()){
                throw new CELFormatException("format err :" + exp);
            }
            args = list.toArray();
        }else{
            args = new Object[0];
        }

        Object o = null;
        try {
            if (methodName != null) {
                o = BeanUtilities.invokeMethod(c, null, methodName, args);
            } else {
                o = BeanUtilities.newInstance(c, args);
            }
        } catch (InvocationBeansException e) {
            throw new CELFormatException("bean invocation err :" + exp, e);
        }
        if (o instanceof Void) {
            throw new CELFormatException("method return type void :" + exp);
        }
        return o;
    }

    /**
     * 式の終端のインデックスを検索して返します。 発見出来ない場合は -1 を返します。
     * 
     * @param binder
     *            処理基のバインダー
     * @param expression
     *            式
     * @param literal
     *            リテラル文字
     * @param begin
     *            開始インデックス
     * @return 式の終端のインデックス
     */
    public int findNextIndex(CELBinder binder, Literal literal, String expression, int begin) {
        int end = (begin + 1);//'@'.length() == 1
        end = Seek.skipTo(expression, ' ', end);
        end = expression.indexOf('(', end);
        if (-1 == end) {
            return -1;
        }
        end++;
        end = Seek.skipTo(expression, ' ', end);
        while (end < expression.length()) {
            end = binder.findNextIndex(expression, end);
            if (-1 == end) {
                return -1;
            }
            end = Seek.skipTo(expression, new char[] { ',', ' ' }, end);
            if (end >= expression.length()) {
                break;
            }
            char c = expression.charAt(end);
            if (c == ')') {
                break;
            }
        }
        end = Seek.skipTo(expression, ' ', end);
        if ((end + 1) < expression.length()) {
            end++;
            char c = expression.charAt(end);
            if (c == '@') {
                return (end + 1);//end to next
            }
        }
        return -1;
    }
}
